import time
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Optional
from uuid import UUID

import humanize
import nh3
import pytz
import shortuuid
from babel.dates import format_datetime
from sqlalchemy.sql import text

from fittrackee import db
from fittrackee.languages import LANGUAGES_DATE_STRING
from fittrackee.users.constants import USER_DATE_FORMAT, USER_TIMEZONE
from fittrackee.users.utils.language import get_language

if TYPE_CHECKING:
    from fittrackee.users.models import User


def get_date_string_for_user(date_to_format: datetime, user: "User") -> str:
    """
    Note: date_to_format is a naive datetime
    """
    user_language = get_language(user.language)
    user_timezone = user.timezone if user.timezone else USER_TIMEZONE
    user_date_format = (
        user.date_format if user.date_format else USER_DATE_FORMAT
    )

    date_format = (
        LANGUAGES_DATE_STRING[user_language]
        if user_date_format == "date_string"
        else user_date_format
    )
    return format_datetime(
        pytz.utc.localize(date_to_format).astimezone(
            pytz.timezone(user_timezone)
        ),
        format=f"{date_format} - HH:mm:ss",
        locale=user_language,
    )


def get_readable_duration(duration: int, locale: Optional[str] = None) -> str:
    """
    Return readable and localized duration from duration in seconds
    """
    if locale is None:
        locale = 'en'
    if locale != 'en':
        try:
            _t = humanize.i18n.activate(locale)  # noqa
        except FileNotFoundError:
            locale = 'en'
    readable_duration = humanize.naturaldelta(timedelta(seconds=duration))
    if locale != 'en':
        humanize.i18n.deactivate()
    return readable_duration


def clean(sql: str, days: int) -> int:
    limit = int(time.time()) - (days * 86400)
    result = db.session.execute(text(sql), {'limit': limit})
    db.session.commit()
    return result.rowcount


def encode_uuid(uuid_value: UUID) -> str:
    """
    Return short id string from a UUID
    """
    return shortuuid.encode(uuid_value)


def decode_short_id(short_id: str) -> UUID:
    """
    Return UUID from a short id string
    """
    return shortuuid.decode(short_id)


def clean_input(text: str) -> str:
    # HTML sanitization
    return nh3.clean(
        text,
        tags={"a", "br", "p", "span"},
        attributes={"a": {"href", "target"}},
    )



from fittrackee.exceptions import GenericException


class CommentForbiddenException(GenericException):
    def __init__(self) -> None:
        super().__init__('error', 'you do not have permissions')



from fittrackee.exceptions import GenericException


class InvalidGPXException(GenericException):
    pass


class WorkoutException(GenericException):
    pass


class WorkoutGPXException(GenericException):
    pass


class WorkoutForbiddenException(GenericException):
    def __init__(self) -> None:
        super().__init__('error', 'you do not have permissions')



class InvalidEquipmentException(Exception):
    def __init__(
        self, status: str, message: str, equipment_short_id: str
    ) -> None:
        super().__init__(message)
        self.status = status
        self.message = message
        self.equipment_id = equipment_short_id


class InvalidEquipmentsException(Exception):
    pass



from datetime import datetime
from typing import Dict, Optional, Union

from sqlalchemy import func

from fittrackee import db
from fittrackee.comments.models import Comment
from fittrackee.comments.utils import get_comment
from fittrackee.users.exceptions import UserNotFoundException
from fittrackee.users.models import User
from fittrackee.users.users_service import UserManagerService
from fittrackee.workouts.models import Workout
from fittrackee.workouts.utils.workouts import get_workout

from .exceptions import (
    InvalidReportActionException,
    InvalidReportException,
    ReportNotFoundException,
    SuspendedObjectException,
    UserWarningExistsException,
)
from .models import (
    ALL_USER_ACTION_TYPES,
    COMMENT_ACTION_TYPES,
    WORKOUT_ACTION_TYPES,
    Report,
    ReportAction,
    ReportActionAppeal,
    ReportComment,
)


class ReportService:
    @staticmethod
    def create_report(
        *,
        reporter: User,
        note: str,
        object_id: str,
        object_type: str,
    ) -> Report:
        if object_type == "comment":
            target_object = get_comment(object_id, reporter)
        elif object_type == "workout":
            target_object = get_workout(object_id, reporter)
        else:  # object_type == "user"
            target_object = User.query.filter(
                func.lower(User.username) == func.lower(object_id),
            ).first()
            if not target_object or not target_object.is_active:
                raise UserNotFoundException()

        if target_object and target_object.suspended_at:
            raise SuspendedObjectException(object_type)

        existing_unresolved_report = Report.query.filter_by(
            reported_by=reporter.id,
            resolved=False,
            object_type=object_type,
            **{f"reported_{object_type}_id": target_object.id},
        ).first()
        if existing_unresolved_report:
            raise InvalidReportException("a report already exists")

        new_report = Report(
            note=note,
            reported_by=reporter.id,
            reported_object=target_object,
        )
        db.session.add(new_report)
        db.session.commit()

        return new_report

    @staticmethod
    def update_report(
        *,
        report_id: int,
        moderator: User,
        report_comment: str,
        resolved: Optional[bool] = None,
    ) -> Report:
        report = Report.query.filter_by(id=report_id).first()
        if not report:
            raise ReportNotFoundException()
        previous_resolved = report.resolved

        new_report_comment = ReportComment(
            comment=report_comment,
            report_id=report_id,
            user_id=moderator.id,
        )
        db.session.add(new_report_comment)

        now = datetime.utcnow()
        report.updated_at = now
        report_action = None
        if resolved is not None:
            report.resolved = resolved
        if resolved is True and report.resolved_by is None:
            report.resolved_at = now
            report.resolved_by = moderator.id
            report_action = ReportAction(
                report_id=report.id,
                moderator_id=moderator.id,
                action_type="report_resolution",
                created_at=now,
            )
        if resolved is False:
            report.resolved_at = None
            report.resolved_by = None
            if previous_resolved is True:
                report_action = ReportAction(
                    report_id=report.id,
                    moderator_id=moderator.id,
                    action_type="report_reopening",
                    created_at=now,
                )

        if report_action:
            db.session.add(report_action)

        db.session.commit()

        return report

    @staticmethod
    def create_report_action(
        *,
        report: Report,
        moderator: User,
        action_type: str,
        reason: Optional[str] = None,
        data: Dict,
    ) -> Optional[ReportAction]:
        reported_user: User = report.reported_user

        # if reported user has been deleted after report creation
        if not reported_user:
            raise InvalidReportActionException("invalid 'username'")

        now = datetime.utcnow()
        report_action = None
        if action_type in ALL_USER_ACTION_TYPES:
            username = data.get("username")
            if not username:
                raise InvalidReportActionException("'username' is missing")
            if username != reported_user.username:
                raise InvalidReportActionException("invalid 'username'")

            if action_type.startswith("user_warning"):
                user = User.query.filter_by(username=username).first()

                existing_report_action = ReportAction.query.filter_by(
                    action_type=action_type,
                    report_id=report.id,
                    user_id=user.id,
                ).first()
                if existing_report_action:
                    raise UserWarningExistsException("user already warned")

                report_action = ReportAction(
                    moderator_id=moderator.id,
                    action_type=action_type,
                    created_at=now,
                    report_id=report.id,
                    reason=reason,
                    user_id=user.id,
                )
                if report.reported_comment_id:
                    report_action.comment_id = report.reported_comment_id
                elif report.reported_workout_id:
                    report_action.workout_id = report.reported_workout_id
                db.session.add(report_action)
            else:
                user_manager_service = UserManagerService(
                    username=username, moderator_id=moderator.id
                )
                user, _, _, _ = user_manager_service.update(
                    suspended=action_type == "user_suspension",
                    report_id=report.id,
                    reason=reason,
                )
                if action_type == "user_unsuspension":
                    appeal = (
                        ReportActionAppeal.query.join(ReportAction)
                        .filter(
                            ReportAction.report_id == report.id,
                            ReportAction.user_id == user.id,
                            ReportAction.action_type == "user_suspension",
                        )
                        .first()
                    )
                    if appeal:
                        appeal.approved = None
                        appeal.updated_at = datetime.utcnow()
                        db.session.flush()

        elif action_type in COMMENT_ACTION_TYPES + WORKOUT_ACTION_TYPES:
            object_type = action_type.split("_")[0]
            object_type_column = f"{object_type}_id"
            object_id = data.get(object_type_column)
            if not object_id:
                raise InvalidReportActionException(
                    f"'{object_type_column}' is missing"
                )
            reported_object: Union[Comment, Workout] = getattr(
                report, f"reported_{object_type}"
            )
            if not reported_object or reported_object.short_id != object_id:
                raise InvalidReportActionException(
                    f"invalid '{object_type_column}'"
                )

            report_action = ReportAction(
                moderator_id=moderator.id,
                action_type=action_type,
                created_at=now,
                report_id=report.id,
                reason=reason,
                user_id=reported_object.user_id,
                **{object_type_column: reported_object.id},
            )
            db.session.add(report_action)

            if "_suspension" in action_type:
                if reported_object.suspended_at:
                    raise InvalidReportActionException(
                        f"{object_type} already suspended"
                    )
                reported_object.suspended_at = now

            else:
                if reported_object.suspended_at is None:
                    raise InvalidReportActionException(
                        f"{object_type} already reactivated"
                    )
                reported_object.suspended_at = None
                appeal = (
                    ReportActionAppeal.query.join(ReportAction)
                    .filter(
                        ReportAction.report_id == report.id,
                        ReportAction.user_id == reported_object.user_id,
                        ReportAction.action_type
                        == f"{object_type}_suspension",
                    )
                    .first()
                )
                if appeal:
                    appeal.approved = None
                    appeal.updated_at = datetime.utcnow()
                    db.session.flush()
            db.session.flush()
        else:
            raise InvalidReportActionException("invalid action type")
        return report_action

    @staticmethod
    def process_appeal(
        appeal: ReportActionAppeal, moderator: User, data: Dict
    ) -> Optional[ReportAction]:
        appeal.moderator_id = moderator.id
        appeal.approved = data["approved"]
        appeal.reason = data["reason"]
        appeal.updated_at = datetime.utcnow()

        action = appeal.action
        content = None
        content_type = ''
        if action.action_type.startswith("comment_"):
            content = Comment.query.filter_by(id=action.comment_id).first()
            content_type = "comment"
        elif action.action_type.startswith("workout_"):
            content = Workout.query.filter_by(id=action.workout_id).first()
            content_type = "workout"

        new_report_action = None
        if data["approved"]:
            if action.action_type == "user_suspension":
                if not appeal.user.suspended_at:
                    raise InvalidReportActionException(
                        "user account has already been reactivated"
                    )

                user_manager_service = UserManagerService(
                    username=appeal.user.username,
                    moderator_id=moderator.id,
                )
                user, _, _, new_report_action = user_manager_service.update(
                    suspended=False, report_id=appeal.action.report_id
                )
            if action.action_type == "user_warning":
                new_report_action = ReportAction(
                    moderator_id=moderator.id,
                    action_type="user_warning_lifting",
                    created_at=datetime.utcnow(),
                    report_id=action.report_id,
                    user_id=action.user_id,
                )
                db.session.add(new_report_action)
            if (
                action.action_type
                in ["comment_suspension", "workout_suspension"]
                and content
            ):
                if not content.suspended_at:
                    raise InvalidReportActionException(
                        f"{content_type} already reactivated"
                    )
                content_id = {f"{content_type}_id": content.id}
                new_report_action = ReportAction(
                    moderator_id=moderator.id,
                    action_type=f"{content_type}_unsuspension",
                    created_at=datetime.utcnow(),
                    report_id=action.report_id,
                    user_id=action.user_id,
                    **content_id,
                )
                db.session.add(new_report_action)
                content.suspended_at = None
        else:
            if (
                action.action_type == "user_suspension"
                and not appeal.user.suspended_at
            ):
                if not appeal.user.suspended_at:
                    raise InvalidReportActionException(
                        "user account has been reactivated after appeal"
                    )
            if (
                action.action_type
                in ["comment_suspension", "workout_suspension"]
                and content
                and not content.suspended_at
            ):
                raise InvalidReportActionException(
                    f"{content_type} has been reactivated after appeal"
                )
        return new_report_action



from datetime import datetime
from typing import Dict, Optional, Union
from uuid import uuid4

from sqlalchemy.dialects import postgresql
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.orm import Mapper, Session

from fittrackee import BaseModel, db
from fittrackee.comments.exceptions import CommentForbiddenException
from fittrackee.comments.models import Comment
from fittrackee.users.models import Notification, User
from fittrackee.users.roles import UserRole
from fittrackee.utils import encode_uuid
from fittrackee.workouts.exceptions import WorkoutForbiddenException
from fittrackee.workouts.models import Workout

from .exceptions import (
    InvalidReportActionAppealException,
    InvalidReportActionAppealUserException,
    InvalidReportActionException,
    InvalidReporterException,
    InvalidReportException,
    ReportActionAppealForbiddenException,
    ReportActionForbiddenException,
    ReportCommentForbiddenException,
    ReportForbiddenException,
)

REPORT_OBJECT_TYPES = [
    "comment",
    "user",
    "workout",
]
REPORT_ACTION_TYPES = [
    "report_reopening",
    "report_resolution",
]
USER_ACTION_TYPES = [
    "user_suspension",
    "user_unsuspension",
    "user_warning",
]
ALL_USER_ACTION_TYPES = USER_ACTION_TYPES + [
    "user_warning_lifting",
]
COMMENT_ACTION_TYPES = [
    "comment_suspension",
    "comment_unsuspension",
]
WORKOUT_ACTION_TYPES = [
    "workout_suspension",
    "workout_unsuspension",
]
OBJECTS_ACTION_TYPES = (
    COMMENT_ACTION_TYPES + ALL_USER_ACTION_TYPES + WORKOUT_ACTION_TYPES
)
ALL_ACTION_TYPES = REPORT_ACTION_TYPES + OBJECTS_ACTION_TYPES


class Report(BaseModel):
    __tablename__ = 'reports'
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    created_at = db.Column(db.DateTime, default=datetime.utcnow)
    updated_at = db.Column(db.DateTime, nullable=True)
    resolved_at = db.Column(db.DateTime, nullable=True)
    reported_by = db.Column(
        db.Integer,
        db.ForeignKey('users.id', ondelete='SET NULL'),
        index=True,
        nullable=True,
    )
    reported_comment_id = db.Column(
        db.Integer,
        db.ForeignKey('comments.id', ondelete='SET NULL'),
        index=True,
        nullable=True,
    )
    reported_user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id', ondelete='SET NULL'),
        index=True,
        nullable=True,
    )
    reported_workout_id = db.Column(
        db.Integer,
        db.ForeignKey('workouts.id', ondelete='SET NULL'),
        index=True,
        nullable=True,
    )
    resolved_by = db.Column(
        db.Integer,
        db.ForeignKey('users.id', ondelete='SET NULL'),
        index=True,
        nullable=True,
    )
    resolved = db.Column(db.Boolean, default=False, nullable=False)
    object_type = db.Column(db.String(50), nullable=False, index=True)
    note = db.Column(db.String(), nullable=False)

    reported_comment = db.relationship(
        'Comment',
        lazy=True,
        backref=db.backref('comment_reports', lazy='select'),
    )
    reported_user = db.relationship(
        'User',
        primaryjoin=reported_user_id == User.id,
        backref=db.backref('user_reports', lazy='select'),
    )
    reported_workout = db.relationship(
        'Workout',
        lazy=True,
        backref=db.backref('workouts_reports', lazy='select'),
    )
    reporter = db.relationship(
        'User',
        primaryjoin=reported_by == User.id,
        backref=db.backref(
            'user_own_reports',
            lazy='select',
            single_parent=True,
        ),
    )
    resolver = db.relationship(
        'User',
        primaryjoin=resolved_by == User.id,
        backref=db.backref(
            'user_resolved_reports',
            lazy='select',
            single_parent=True,
        ),
    )
    comments = db.relationship(
        'ReportComment',
        backref=db.backref('report', lazy='select'),
    )
    report_actions = db.relationship(
        'ReportAction',
        lazy=True,
        backref=db.backref('report', lazy='select', single_parent=True),
        order_by='ReportAction.created_at.asc()',
    )

    @property
    def reported_object(self) -> Union[Comment, None, User, Workout]:
        # util method, used by tests
        if self.object_type == "comment":
            return self.reported_comment
        if self.object_type == "user":
            return self.reported_user
        if self.object_type == "workout":
            return self.reported_workout
        return None

    @property
    def is_reported_user_warned(self) -> bool:
        return (
            ReportAction.query.filter_by(
                action_type="user_warning",
                report_id=self.id,
                user_id=self.reported_user_id,
            ).first()
            is not None
        )

    def __init__(
        self,
        note: str,
        reported_by: int,
        reported_object: Union[Comment, User, Workout],
        created_at: Optional[datetime] = None,
    ):
        object_type = reported_object.__class__.__name__.lower()
        if object_type not in REPORT_OBJECT_TYPES:
            raise InvalidReportException()
        user_id = (
            reported_object.id
            if object_type == "user"
            else reported_object.user_id
        )
        if user_id == reported_by:
            raise InvalidReporterException()

        self.created_at = created_at if created_at else datetime.utcnow()
        self.note = note
        self.object_type = object_type
        self.reported_by = reported_by
        self.reported_comment_id = (
            reported_object.id if object_type == "comment" else None
        )
        self.reported_user_id = user_id
        self.reported_workout_id = (
            reported_object.id if object_type == "workout" else None
        )
        self.resolved = False

    def serialize(self, current_user: User, full: bool = False) -> Dict:
        if (
            not current_user.has_moderator_rights
            and self.reported_by != current_user.id
        ):
            raise ReportForbiddenException()

        try:
            reported_comment = (
                self.reported_comment.serialize(current_user, for_report=True)
                if self.reported_comment_id
                else None
            )
        except CommentForbiddenException:
            reported_comment = '_COMMENT_UNAVAILABLE_'

        try:
            reported_workout = (
                self.reported_workout.serialize(
                    user=current_user, for_report=True
                )
                if self.reported_workout
                else None
            )
        except WorkoutForbiddenException:
            reported_workout = '_WORKOUT_UNAVAILABLE_'

        report = {
            "created_at": self.created_at,
            "id": self.id,
            "is_reported_user_warned": self.is_reported_user_warned,
            "note": self.note,
            "object_type": self.object_type,
            "reported_by": (
                self.reporter.serialize(current_user=current_user)
                if self.reported_by
                else None
            ),
            "reported_comment": reported_comment,
            "reported_user": (
                self.reported_user.serialize(current_user=current_user)
                if self.reported_user_id
                else None
            ),
            "reported_workout": reported_workout,
            "resolved": self.resolved,
            "resolved_at": self.resolved_at,
        }
        if current_user.has_moderator_rights:
            if full:
                report["report_actions"] = [
                    action.serialize(current_user, full=False)
                    for action in self.report_actions
                ]
                report["comments"] = [
                    comment.serialize(current_user)
                    for comment in self.comments
                ]
            report["resolved_by"] = (
                None
                if self.resolved_by is None
                else self.resolver.serialize(current_user=current_user)
            )
            report["updated_at"] = self.updated_at
        return report


@listens_for(Report, 'after_insert')
def on_report_insert(
    mapper: Mapper, connection: Connection, new_report: Report
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        from fittrackee.users.models import Notification, User

        for admin in User.query.filter(
            User.role >= UserRole.MODERATOR.value,
            User.id != new_report.reported_by,
            User.is_active == True,  # noqa
        ).all():
            notification = Notification(
                from_user_id=new_report.reported_by,
                to_user_id=admin.id,
                created_at=new_report.created_at,
                event_type='report',
                event_object_id=new_report.id,
            )
            session.add(notification)


class ReportComment(BaseModel):
    __tablename__ = 'report_comments'
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    created_at = db.Column(db.DateTime, default=datetime.utcnow)
    report_id = db.Column(
        db.Integer,
        db.ForeignKey('reports.id', ondelete='CASCADE'),
        index=True,
        nullable=True,
    )
    user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id', ondelete='CASCADE'),
        index=True,
        nullable=True,
    )
    comment = db.Column(db.String(), nullable=False)

    user = db.relationship(
        'User',
        backref=db.backref('user_report_comments', lazy='select'),
    )

    def __init__(
        self,
        report_id: int,
        user_id: int,
        comment: str,
        created_at: Optional[datetime] = None,
    ):
        self.created_at = created_at if created_at else datetime.utcnow()
        self.comment = comment
        self.report_id = report_id
        self.user_id = user_id

    def serialize(self, current_user: User) -> Dict:
        if not current_user.has_moderator_rights:
            raise ReportCommentForbiddenException()
        return {
            "created_at": self.created_at,
            "comment": self.comment,
            "id": self.id,
            "report_id": self.report_id,
            "user": self.user.serialize(current_user=current_user),
        }


class ReportAction(BaseModel):
    __tablename__ = "report_actions"
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    uuid = db.Column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    created_at = db.Column(db.DateTime, default=datetime.utcnow, index=True)
    moderator_id = db.Column(
        db.Integer,
        db.ForeignKey("users.id", ondelete="SET NULL"),
        index=True,
        nullable=True,  # to keep log if an admin is deleted
    )
    report_id = db.Column(
        db.Integer,
        db.ForeignKey("reports.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    user_id = db.Column(
        db.Integer,
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
        nullable=True,
    )
    workout_id = db.Column(
        db.Integer,
        db.ForeignKey('workouts.id', ondelete='SET NULL'),
        index=True,
        nullable=True,
    )
    comment_id = db.Column(
        db.Integer,
        db.ForeignKey('comments.id', ondelete='SET NULL'),
        index=True,
        nullable=True,
    )
    action_type = db.Column(db.String(50), nullable=False)
    reason = db.Column(db.String(), nullable=True)

    moderator = db.relationship(
        "User",
        primaryjoin=moderator_id == User.id,
        lazy="joined",
        single_parent=True,
    )
    user = db.relationship(
        "User",
        primaryjoin=user_id == User.id,
        lazy="joined",
        single_parent=True,
    )
    appeal = db.relationship(
        "ReportActionAppeal",
        uselist=False,
        backref=db.backref("action", lazy='joined', single_parent=True),
    )
    comment = db.relationship(
        'Comment',
        lazy=True,
        backref=db.backref('comment_report_action', lazy='select'),
    )
    workout = db.relationship(
        'Workout',
        lazy=True,
        backref=db.backref('workout_report_action', lazy='select'),
    )

    def __init__(
        self,
        action_type: str,
        moderator_id: int,
        report_id: int,
        *,
        user_id: Optional[int] = None,
        comment_id: Optional[int] = None,
        workout_id: Optional[int] = None,
        reason: Optional[str] = None,
        created_at: Optional[datetime] = None,
    ):
        if action_type not in ALL_ACTION_TYPES:
            raise InvalidReportActionException()
        if action_type in OBJECTS_ACTION_TYPES and not user_id:
            raise InvalidReportActionException()
        if action_type in WORKOUT_ACTION_TYPES and not workout_id:
            raise InvalidReportActionException()
        if action_type in COMMENT_ACTION_TYPES and not comment_id:
            raise InvalidReportActionException()

        self.action_type = action_type
        self.moderator_id = moderator_id
        self.created_at = created_at if created_at else datetime.utcnow()
        self.comment_id = (
            comment_id if action_type in COMMENT_ACTION_TYPES else None
        )
        self.reason = reason
        self.report_id = report_id
        self.user_id = user_id if action_type in OBJECTS_ACTION_TYPES else None
        self.workout_id = (
            workout_id if action_type in WORKOUT_ACTION_TYPES else None
        )

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    def serialize(self, current_user: User, full: bool = True) -> Dict:
        if (
            not current_user.has_moderator_rights
            and current_user.id != self.user_id
        ):
            raise ReportActionForbiddenException()
        action = {
            "action_type": self.action_type,
            "appeal": (
                self.appeal.serialize(current_user) if self.appeal else None
            ),
            "created_at": self.created_at,
            "id": self.short_id,
            "reason": self.reason,
        }

        if current_user.has_moderator_rights:
            action["report_id"] = self.report_id
            action["moderator"] = self.moderator.serialize(
                current_user=current_user
            )
            action["user"] = (
                self.user.serialize(current_user=current_user)
                if self.user
                else None
            )
        if not full:
            return action

        if current_user.has_moderator_rights:
            action = {
                **action,
                "comment": (
                    self.comment.serialize(user=current_user, for_report=True)
                    if self.comment_id
                    else None
                ),
                "workout": (
                    self.workout.serialize(user=current_user, for_report=True)
                    if self.workout_id
                    else None
                ),
            }
        else:
            action["comment"] = (
                self.comment.serialize(user=current_user)
                if self.comment_id
                else None
            )
            action["workout"] = (
                self.workout.serialize(user=current_user)
                if self.workout_id
                else None
            )

        return action


class ReportActionAppeal(BaseModel):
    __tablename__ = "report_action_appeals"
    __table_args__ = (
        db.UniqueConstraint(
            'action_id', 'user_id', name='action_id_user_id_unique'
        ),
    )

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    uuid = db.Column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    action_id = db.Column(
        db.Integer,
        db.ForeignKey("report_actions.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    user_id = db.Column(
        db.Integer,
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    moderator_id = db.Column(
        db.Integer,
        db.ForeignKey("users.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    created_at = db.Column(
        db.DateTime, default=datetime.utcnow, nullable=False
    )
    updated_at = db.Column(db.DateTime)
    approved = db.Column(db.Boolean, nullable=True)
    text = db.Column(db.String(), nullable=False)
    reason = db.Column(db.String(), nullable=True)

    moderator = db.relationship(
        "User",
        primaryjoin=moderator_id == User.id,
        lazy="joined",
        single_parent=True,
    )
    user = db.relationship(
        "User",
        primaryjoin=user_id == User.id,
        lazy="joined",
        single_parent=True,
    )

    def __init__(
        self,
        action_id: int,
        user_id: int,
        text: str,
        created_at: Optional[datetime] = None,
    ):
        action = ReportAction.query.filter_by(id=action_id).first()
        if action.action_type not in [
            "comment_suspension",
            "user_suspension",
            "user_warning",
            "workout_suspension",
        ]:
            raise InvalidReportActionAppealException()
        if action.user_id != user_id:
            raise InvalidReportActionAppealUserException()
        self.action_id = action_id
        self.created_at = created_at if created_at else datetime.utcnow()
        self.text = text
        self.user_id = user_id

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    def serialize(self, current_user: User) -> Dict:
        if (
            not current_user.has_moderator_rights
            and current_user.id != self.user_id
        ):
            raise ReportActionAppealForbiddenException()
        appeal = {
            "approved": self.approved,
            "created_at": self.created_at,
            "id": self.short_id,
            "reason": self.reason,
            "text": self.text,
            "updated_at": self.updated_at,
        }
        if current_user.has_moderator_rights:
            appeal["moderator"] = (
                self.moderator.serialize(current_user=current_user)
                if self.moderator
                else None
            )
            appeal["user"] = self.user.serialize(current_user=current_user)
        return appeal


@listens_for(ReportAction, 'after_insert')
def on_report_action_insert(
    mapper: Mapper, connection: Connection, new_action: ReportAction
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        if (
            new_action.action_type
            in COMMENT_ACTION_TYPES
            + WORKOUT_ACTION_TYPES
            + ["user_warning", "user_warning_lifting"]
        ):
            notification = Notification(
                from_user_id=new_action.moderator_id,
                to_user_id=new_action.user_id,
                created_at=new_action.created_at,
                event_type=new_action.action_type,
                event_object_id=new_action.id,
            )
            session.add(notification)


@listens_for(ReportActionAppeal, 'after_insert')
def on_report_action_appeal_insert(
    mapper: Mapper, connection: Connection, new_appeal: ReportActionAppeal
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        from fittrackee.users.models import Notification, User

        report_action = ReportAction.query.filter_by(
            id=new_appeal.action_id
        ).first()
        if report_action:
            for admin in User.query.filter(
                User.role >= UserRole.MODERATOR.value,
                User.id != new_appeal.user_id,
                User.is_active == True,  # noqa
            ).all():
                notification = Notification(
                    from_user_id=new_appeal.user_id,
                    to_user_id=admin.id,
                    created_at=new_appeal.created_at,
                    event_type=(
                        'user_warning_appeal'
                        if report_action.action_type == 'user_warning'
                        else 'suspension_appeal'
                    ),
                    event_object_id=new_appeal.id,
                )
                session.add(notification)



LANGUAGES_DATE_STRING = {
    "bg": "d MMM yyyy",  # Bulgarian
    "cs": "d. MMM yyyy",  # Czech
    "de": "d. MMM yyyy",  # German
    "en": "MMM. d, yyyy",  # English
    "es": "d MMM yyyy",  # Spanish
    "eu": "yyyy MMM d",  # Basque
    "fr": "d MMM yyyy",  # French
    "gl": "d MMM yyyy",  # Galician
    "it": "d MMM yyyy",  # Italian
    "nb": "d. MMM yyyy",  # Norwegian Bokmål
    "nl": "d MMM. yyyy",  # Dutch
    "pl": "d MMM yyyy",  # Polish
    "pt": "d MMM yyyy",  # Portuguese
    "ru": "d MMM yyyy",  # Russian
}

SUPPORTED_LANGUAGES = list(LANGUAGES_DATE_STRING.keys())



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple
from uuid import uuid4

from sqlalchemy.dialects import postgresql
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import select, text
from sqlalchemy.types import Enum

from fittrackee import BaseModel, db
from fittrackee.utils import encode_uuid
from fittrackee.visibility_levels import VisibilityLevel, can_view

from .exceptions import CommentForbiddenException

if TYPE_CHECKING:
    from fittrackee.reports.models import ReportAction
    from fittrackee.users.models import User


def get_comments(workout_id: int, user: Optional['User']) -> List['Comment']:
    if user:
        params = {"workout_id": workout_id, "user_id": user.id}
        sql = """
        SELECT comments.*
        FROM comments
        LEFT OUTER JOIN mentions ON mentions.comment_id = comments.id
        WHERE comments.workout_id = :workout_id
          AND comments.user_id NOT IN (
            SELECT blocked_users.user_id
            FROM blocked_users
            WHERE blocked_users.by_user_id = :user_id
          )
          AND comments.user_id NOT IN (
            SELECT blocked_users.by_user_id
            FROM blocked_users
            WHERE blocked_users.user_id = user_id
          )
          AND (comments.user_id = :user_id
            OR (
              mentions.user_id = :user_id
              OR comments.text_visibility = 'PUBLIC'
              OR (comments.text_visibility = 'FOLLOWERS' AND :user_id IN (
                SELECT follower_user_id
                FROM follow_requests
                WHERE follower_user_id = :user_id
                  AND followed_user_id = comments.user_id
                  AND is_approved IS TRUE
              ))
            )
          )
        ORDER BY comments.created_at;"""

        comments_filter = db.session.scalars(
            select(Comment)
            .from_statement(
                text(sql),
            )
            .params(**params)
        ).unique()
    else:
        comments_filter = Comment.query.filter(
            Comment.workout_id == workout_id,
            Comment.text_visibility == VisibilityLevel.PUBLIC,
        ).order_by(Comment.created_at.asc())

    return comments_filter.all()


class Comment(BaseModel):
    __tablename__ = 'comments'
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    uuid = db.Column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id', ondelete='CASCADE'),
        index=True,
        nullable=False,
    )
    workout_id = db.Column(
        db.Integer,
        db.ForeignKey('workouts.id', ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    created_at = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    modification_date = db.Column(db.DateTime, nullable=True)
    text = db.Column(db.String(), nullable=False)
    text_visibility = db.Column(
        Enum(VisibilityLevel, name='visibility_levels'),
        server_default='PRIVATE',
        nullable=False,
    )
    suspended_at = db.Column(db.DateTime, nullable=True)

    mentions = db.relationship(
        "Mention",
        lazy=True,
        cascade="all, delete",
        backref=db.backref("comment", lazy="joined", single_parent=True),
    )
    mentioned_users = db.relationship(
        "User",
        secondary="mentions",
        primaryjoin="Comment.id == Mention.comment_id",
        secondaryjoin="Mention.user_id == User.id",
        lazy="dynamic",
        viewonly=True,
    )
    likes = db.relationship(
        "User",
        secondary="comment_likes",
        primaryjoin="Comment.id == CommentLike.comment_id",
        secondaryjoin="CommentLike.user_id == User.id",
        lazy="dynamic",
        viewonly=True,
    )

    def __repr__(self) -> str:
        return f'<Comment {self.id}>'

    def __init__(
        self,
        user_id: int,
        workout_id: int,
        text: str,
        text_visibility: VisibilityLevel,
        created_at: Optional[datetime.datetime] = None,
    ) -> None:
        self.user_id = user_id
        self.workout_id = workout_id
        self.text = text
        self.text_visibility = text_visibility
        self.created_at = (
            datetime.datetime.utcnow() if created_at is None else created_at
        )

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    @property
    def suspension_action(self) -> Optional['ReportAction']:
        if self.suspended_at is None:
            return None

        from fittrackee.reports.models import ReportAction

        return (
            ReportAction.query.filter(
                ReportAction.comment_id == self.id,
                ReportAction.action_type == "comment_suspension",
            )
            .order_by(ReportAction.created_at.desc())
            .first()
        )

    def handle_mentions(self) -> Tuple[str, Set['User']]:
        from .utils import handle_mentions

        return handle_mentions(self.text)

    def create_mentions(self) -> Tuple[str, Set['User']]:
        linkified_text, mentioned_users = self.handle_mentions()
        for user in mentioned_users:
            mention = Mention(comment_id=self.id, user_id=user.id)
            db.session.add(mention)
            db.session.flush()
        return linkified_text, mentioned_users

    def update_mentions(self) -> None:
        from fittrackee.users.models import Notification, User

        existing_mentioned_users = set(
            db.session.query(User)
            .join(Mention, User.id == Mention.user_id)
            .all()
        )
        linkified_text, updated_mentioned_users = self.handle_mentions()
        unchanged_mentions = updated_mentioned_users.intersection(
            existing_mentioned_users
        )

        # delete removed mentions
        deleted_mentioned_users = existing_mentioned_users - unchanged_mentions
        mentions_to_delete = {user.id for user in deleted_mentioned_users}
        Mention.query.filter(
            Mention.comment_id == self.id,
            Mention.user_id.in_(mentions_to_delete),
        ).delete()
        Notification.query.filter(
            Notification.to_user_id.in_(mentions_to_delete),
            Notification.event_type == 'mention',
            Notification.event_object_id == self.id,
        ).delete()
        db.session.flush()

        # create new mentions
        for user in updated_mentioned_users - unchanged_mentions:
            mention = Mention(comment_id=self.id, user_id=user.id)
            db.session.add(mention)
        db.session.flush()

    def liked_by(self, user: 'User') -> bool:
        return user in self.likes.all()

    def serialize(
        self,
        user: Optional['User'] = None,
        for_report: bool = False,
    ) -> Dict:
        if not can_view(self, 'text_visibility', user, for_report):
            raise CommentForbiddenException

        # suspended comment content is only visible to its owner or
        # to admin in report only
        suspension: Dict[str, Any] = {}
        if self.suspended_at:
            suspension["suspended"] = True
            if user and user.id == self.user_id and self.suspension_action:
                suspension["suspension"] = self.suspension_action.serialize(
                    current_user=user, full=False
                )
        if user and (
            user.id == self.user_id
            or (user.has_moderator_rights and for_report)
        ):
            suspension["suspended_at"] = self.suspended_at

        display_content = (
            False
            if (
                self.suspended_at
                and (
                    not user
                    or (user.has_moderator_rights and not for_report)
                    or (
                        not (user.has_moderator_rights and for_report)
                        and user.id != self.user_id
                    )
                )
            )
            else True
        )

        return {
            'id': self.short_id,
            'user': self.user.serialize(),
            'workout_id': (
                self.workout.short_id
                if self.workout
                and can_view(self.workout, 'workout_visibility', user)
                else None
            ),
            'text': self.text if display_content else None,
            'text_html': (
                self.handle_mentions()[0] if display_content else None
            ),
            'text_visibility': self.text_visibility,
            'created_at': self.created_at,
            'modification_date': self.modification_date,
            'mentions': (
                [
                    mentioned_user.serialize()
                    for mentioned_user in self.mentioned_users
                ]
                if display_content
                else []
            ),
            'likes_count': self.likes.count() if display_content else 0,
            'liked': self.liked_by(user) if user else False,
            **suspension,
        }


class Mention(BaseModel):
    __tablename__ = 'mentions'

    comment_id = db.Column(
        db.Integer,
        db.ForeignKey('comments.id', ondelete="CASCADE"),
        primary_key=True,
    )
    user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id', ondelete="CASCADE"),
        primary_key=True,
    )
    created_at = db.Column(
        db.DateTime, nullable=False, default=datetime.datetime.utcnow
    )

    def __init__(
        self,
        comment_id: int,
        user_id: int,
        created_at: Optional[datetime.datetime] = None,
    ):
        self.comment_id = comment_id
        self.user_id = user_id
        self.created_at = (
            datetime.datetime.utcnow() if created_at is None else created_at
        )


@listens_for(Comment, 'after_insert')
def on_comment_insert(
    mapper: Mapper, connection: Connection, new_comment: Comment
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        from fittrackee.users.models import FollowRequest, Notification
        from fittrackee.workouts.models import Workout

        # it creates notification on comment creation only when:
        # - the comment author (notification.from_user_id) is not the recipient
        #   (notification.to_user)
        # - the comment is public
        # - the recipient follows the comment author if privacy level is only
        #   followers

        create_notification = False
        if new_comment.text_visibility == VisibilityLevel.PUBLIC:
            create_notification = True

        workout = Workout.query.filter_by(id=new_comment.workout_id).first()
        to_user_id = workout.user_id

        if new_comment.user_id == to_user_id:
            return

        if (
            not create_notification
            and new_comment.text_visibility == VisibilityLevel.FOLLOWERS
        ):
            create_notification = (
                FollowRequest.query.filter_by(
                    follower_user_id=to_user_id,
                    followed_user_id=new_comment.user_id,
                    is_approved=True,
                ).first()
                is not None
            )

        if not create_notification:
            return

        notification = Notification(
            from_user_id=new_comment.user_id,
            to_user_id=to_user_id,
            created_at=new_comment.created_at,
            event_type='workout_comment',
            event_object_id=new_comment.id,
        )
        session.add(notification)


@listens_for(Comment, 'after_delete')
def on_comment_delete(
    mapper: Mapper, connection: Connection, old_comment: Comment
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        from fittrackee.users.models import Notification

        # delete all notifications related to deleted comment
        Notification.query.filter_by(
            event_object_id=old_comment.id,
        ).delete()


@listens_for(Mention, 'after_insert')
def on_mention_insert(
    mapper: Mapper, connection: Connection, new_mention: Mention
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        from fittrackee.users.models import Notification

        comment = Comment.query.filter_by(id=new_mention.comment_id).first()

        # `mention` notification is not created when:

        # - mentioned user is comment author
        if new_mention.user_id == comment.user_id:
            return

        # - mentioned user is workout owner and `workout_comment'
        # notification does not exist
        notification = (
            Notification.query.join(
                Comment, Comment.id == Notification.event_object_id
            )
            .filter(
                Comment.id == comment.id,
                Notification.event_type == 'workout_comment',
                Notification.to_user_id == new_mention.user_id,
            )
            .first()
        )
        if notification:
            return

        notification = Notification(
            from_user_id=comment.user_id,
            to_user_id=new_mention.user_id,
            created_at=new_mention.created_at,
            event_type='mention',
            event_object_id=comment.id,
        )
        session.add(notification)


@listens_for(Mention, 'after_delete')
def on_mention_delete(
    mapper: Mapper, connection: Connection, old_mention: Mention
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        from fittrackee.users.models import Notification

        Notification.query.filter_by(
            to_user_id=old_mention.user_id,
            event_type='mention',
            event_object_id=old_mention.comment_id,
        ).delete()


class CommentLike(BaseModel):
    __tablename__ = 'comment_likes'
    __table_args__ = (
        db.UniqueConstraint(
            'user_id', 'comment_id', name='user_id_comment_id_unique'
        ),
    )
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    created_at = db.Column(db.DateTime, nullable=False)
    user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id', ondelete='CASCADE'),
        nullable=False,
    )
    comment_id = db.Column(
        db.Integer,
        db.ForeignKey('comments.id', ondelete="CASCADE"),
        nullable=False,
    )

    user = db.relationship("User", lazy=True)
    comment = db.relationship("Comment", lazy=True)

    def __init__(
        self,
        user_id: int,
        comment_id: int,
        created_at: Optional[datetime.datetime] = None,
    ) -> None:
        self.user_id = user_id
        self.comment_id = comment_id
        self.created_at = (
            datetime.datetime.utcnow() if created_at is None else created_at
        )


@listens_for(CommentLike, 'after_insert')
def on_comment_like_insert(
    mapper: Mapper, connection: Connection, new_comment_like: CommentLike
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        from fittrackee.users.models import Notification

        comment = Comment.query.filter_by(
            id=new_comment_like.comment_id
        ).first()
        if new_comment_like.user_id != comment.user_id:
            notification = Notification(
                from_user_id=new_comment_like.user_id,
                to_user_id=comment.user_id,
                created_at=new_comment_like.created_at,
                event_type='comment_like',
                event_object_id=comment.id,
            )
            session.add(notification)


@listens_for(CommentLike, 'after_delete')
def on_comment_like_delete(
    mapper: Mapper, connection: Connection, old_comment_like: CommentLike
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        from fittrackee.users.models import Notification

        comment = Comment.query.filter_by(
            id=old_comment_like.comment_id
        ).first()
        Notification.query.filter_by(
            from_user_id=old_comment_like.user_id,
            to_user_id=comment.user_id,
            event_type='comment_like',
            event_object_id=comment.id,
        ).delete()
